/*
 * Copyright (C) 1996-2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "config.h"
__RCSID("$LAAS$");

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>

#include <GL/glu.h>
#include <tcl.h>
#include <togl.h>

#include "gdhe.h"
#include "global.h"
#include "cb.h"
#include "temps.h"

static float mat_shininess[] = { 100.0 };

#ifndef DEG_TO_RAD
#define DEG_TO_RAD(x) ((x)*M_PI/180.)
#endif

/*----------------------------------------------------------------------*/

/**
 ** create_cb - called when a Togl widget is created
 **
 ** sets the default parameters for this window
 **/
void
create_cb(const struct Togl *togl)
{
    Tcl_HashEntry *entry;
    Tcl_Interp *interp = Togl_Interp(togl);
    int newEntry;
    GDHE_OBS *obs;

    /* font */
    if (FontBase == 0) {
	FontBase = Togl_LoadBitmapFont(togl, TOGL_BITMAP_8_BY_13);
	if (!FontBase) {
	    printf("Couldn't load font!\n");
	    exit(1);
	}
    }
    /* default materials */
    glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
    /* default lighting and shading */
    glDepthFunc(GL_LEQUAL);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);

    glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);

    /* Find an entry in the observer h-table */
    entry  = Tcl_FindHashEntry(posTable, Togl_Ident(togl));
    if (entry == NULL) {
	/* make new observer */
	obs = (GDHE_OBS *)Tcl_Alloc(sizeof(GDHE_OBS));

	/* Get default values from Tcl interpreter, if defined */
	gdheGetDefaultFromInterp(interp, "::obsDist", &obs->dist, 3.0);
	gdheGetDefaultFromInterp(interp, "::obsElev", &obs->elev, 25.0);
	gdheGetDefaultFromInterp(interp, "::obsAzi", &obs->azi, 0.0);
	gdheGetDefaultFromInterp(interp, "::obsX", &obs->x, 0.0);
	gdheGetDefaultFromInterp(interp, "::obsY", &obs->y, 0.0);
	gdheGetDefaultFromInterp(interp, "::obsZ", &obs->z, 0.0);

	/* Initial viewing mode */
	obs->viewMatMode = 0;
	obs->projMode = PERSPECTIVE;
	obs->fovy = 60;
	obs->renderMode = GL_RENDER;
	/* add in the h-table */
	entry = Tcl_CreateHashEntry(posTable, Togl_Ident(togl),
				    &newEntry);
	Tcl_SetHashValue(entry, (ClientData)obs);
    }
}

/*----------------------------------------------------------------------*/

/**
 ** reshape_cb - called after the OpenGL window was resized
 **
 ** recomputes the projection matrix (let OpenGL do it)
 **/
void
reshape_cb(const struct Togl *togl)
{
    int width, height;
    GLfloat aspect;

    width = Togl_Width( togl );
    height = Togl_Height( togl );
    aspect = (float) width / (float) height;

    /* new viewport dimensions */
    glViewport( 0, 0, width, height );
}

/*----------------------------------------------------------------------*/

/*
 * Display a character string
 */
static void
print_string(const char *s)
{
    glCallLists(strlen(s), GL_UNSIGNED_BYTE, s);
}

/*----------------------------------------------------------------------*/

/**
 ** display_cb - called to redisplay a Togl widget
 **
 ** Set the viewing transform and call the Tcl 'draw_all'
 ** procedure to do the dirty job
 **/
void
display_cb(const struct Togl *togl)
{
    double maxdist;
    Tcl_Interp *interp = Togl_Interp(togl);
    Tcl_HashEntry *entry;
    GDHE_OBS *obs;
    char buf[1024];
    struct timeval debut, fin;
    unsigned long duree;
    int width, height;
    GLfloat aspect;
    GLint viewport[4];

    /* record start time  */
    gettimeofday(&debut, NULL);

    /* Clear display */

    glClearColor(backgroundColor[0], backgroundColor[1], backgroundColor[2],
		 backgroundColor[3]);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    /* find the observer in the h-table */

    entry  = Tcl_FindHashEntry(posTable, Togl_Ident(togl));
    if (entry == NULL) {
	 fprintf(stderr, "Can't find a Togl position\n");
	 return;
    }
    obs = (GDHE_OBS *)Tcl_GetHashValue(entry);
    width = Togl_Width( togl );
    height = Togl_Height( togl );
    aspect = (float) width / (float) height;

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    if (obs->renderMode == GL_SELECT) {
	    glGetIntegerv(GL_VIEWPORT, viewport);
	    gluPickMatrix((GLdouble)obs->pickX, 
		(GLdouble)(viewport[3] - obs->pickY), 5.0, 5.0, viewport);
    }
    /* projection du monde */
    switch (obs->projMode) {
    case CUSTOM:
	 glMultMatrixd(obs->projMat);
	 break;
    case PERSPECTIVE:
	 maxdist = hypot(xmax - xmin, ymax - ymin)*3;
	 gluPerspective(obs->fovy, aspect, 0.01, maxdist);

	 break;
    case ORTHOGONAL:
	 maxdist = hypot(xmax - xmin, ymax - ymin)*2;
	 glOrtho(-obs->dist*aspect, obs->dist*aspect,
		 -obs->dist, obs->dist,  0.01, maxdist);

	 break;
    }

    /* Set the position to Look at the world  */
    switch (obs->viewMatMode) {
    case 0:
	    glTranslated(0.0, 0.0, -obs->dist);
	    glRotated(obs->elev, 1.0, 0.0, 0.0);
	    glRotated(obs->azi, 0.0, 1.0, 0.0);
	    /* Changement de repere GL -> LAAS */
	    glRotated(-90, 1, 0, 0);
	    /* translation */
	    glTranslated(-obs->x, -obs->y, -obs->z);
	    break;
    case 1:
	    glMultMatrixd(obs->viewMat);
	    break;
    case 2:
	    /* XXX  undocumented hack to set the observer globally */
	    glRotated(90-obs->azi, 0.0, 0.0, 1.0);
	    glRotated(-obs->elev, 0.0, 1.0, 0.0);
	    glRotated(-obs->dist, 1.0, 0.0, 0.0); /* dist is redefined as the
					   3d rotation */
	    glTranslated(-obs->x, -obs->y, -obs->z);
	    break;
    }

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    if (obs->renderMode == GL_RENDER) {
	    /* lights */
	    if (enable_lighting) {
		    glEnable(GL_LIGHTING);
	    } else {
		    glDisable(GL_LIGHTING);
	    }
	    glEnable( GL_DEPTH_TEST );
    }

    /* Call the Tcl draw_all function */
    snprintf(buf, sizeof(buf), "draw_all %s", Tk_PathName(Togl_TkWin(togl)));
    if (Tcl_Eval(interp, buf) != TCL_OK) {
	Tcl_BackgroundError(interp);
    }

    if (obs->renderMode == GL_RENDER) {
	    /* Text display */
	    glMatrixMode(GL_PROJECTION);
	    glLoadIdentity();
	    glOrtho(-0.5,width-0.5,-0.5,height-0.5,-1.0,1.0);
	    glMatrixMode(GL_MODELVIEW);
	    glLoadIdentity();
	    
	    glDisable( GL_DEPTH_TEST );
	    glDisable( GL_LIGHTING );
	    glLoadIdentity();
	    glColor3f(1.0, 1.0, 1.0);
	    glRasterPos2i(annotateX, annotateY);
	    glListBase(FontBase);
	    
	    print_string(annotateText);

	    Togl_SwapBuffers( togl );

	    /*
	     * display statistics
	     */
	    gettimeofday(&fin, NULL);
	    duree = time_difference(&fin, &debut);
	    snprintf(buf, sizeof(buf), "%ld", duree);
	    Tcl_SetVar(interp, "displayTime", buf, TCL_GLOBAL_ONLY);
	    
	    /* set the date of the next redisplay */
	    schedule_redisplay(&fin, duree);
    }
}

/*----------------------------------------------------------------------*/

/**
 ** Handler des erreurs X
 **/
#ifdef notyet
int
XErrorProc(ClientDate data, XErrorEvent *errEventPtr)
{
    Tk_Window w = (Tk_Window)data;

    fprintf(stderr, "GDHE: X protocol Error: ");
    fprintf(stderr, "error=%d request=%d minor=%d\n",
	    errEventPtr->error_code, errEventPtr->request_code,
	    errEventPtr->minor_code);
    return 0;
}
#endif
